/*
 * arping.c - Ping hosts by ARP requests/replies
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 * Author:	Alexey Kuznetsov <kuznet@ms2.inr.ac.ru>
 * Busybox port: Nick Fedchik <nick@fedchik.org.ua>
 */
/*!
 *******************************************************************************
 * Copyright 2000-2004 Vivotek, Inc. All rights reserved.
 *
 * \file
 * autoipapp.c
 *
 * \brief
 * automatic ip search application for network
 *
 * \date
 * 2004/09/07
 *
 * \author
 * Cheng-Ching Huang
 *
 * \note
 * This program is a modification of BusyBox's arping
 *
 *******************************************************************************
 */

#include <sys/ioctl.h>
#include <sys/signal.h>
#include <sys/time.h>
#include <time.h>

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <arpa/inet.h>
#include <net/if.h>
#include <netinet/ether.h>
#include <netpacket/packet.h>

static struct in_addr src;
static struct in_addr dst;
static struct sockaddr_ll me;
static struct sockaddr_ll he;
static struct timespec last;
static int dad;
static int unsolicited;
static int advert;
static int quiet;
static int quit_on_reply = 0;
static int count = -1;
static int timeout = 4;
static int unicasting;
static int s;
static int broadcast_only;
static int sent;
static int brd_sent;
static int received;
static int brd_recv;
static int req_recv;
static struct timespec start;
unsigned long g_defaultIP;
unsigned long g_seed=0x63;

#define MS_TDIFF(tv1,tv2) ( ((tv1).tv_sec-(tv2).tv_sec)*1000 + \
			   ((tv1).tv_nsec-(tv2).tv_nsec)/1000000 )

void catcher(void);

void usage(void)
{
	printf("Auto IP by pinging hosts by ARP requests/replies.\n"
		   "Usage:\n"
		   "    autoip [-q] [-w timeout] [-I interface] [-h] target\n"
		   "Options:\n"
		   "    -q            Be quiet\n"
		   "    -w timeout    Time to wait for ARP reply each IP, in seconds, default is 4\n"
		   "    -I interface  Outgoing interface name, default is eth0\n"
		   "    -s seed       seed to produce next IP, default is 0x0063, xxx.xxx.0.99\n"
		   "    target        Starting IP address of auto IP\n"
		   "    -h            This help\n");
	exit(1);
}

static int send_pack(int sock, struct in_addr *src_addr,
					 struct in_addr *dst_addr, struct sockaddr_ll *ME,
					 struct sockaddr_ll *HE)
{
	int err;
	struct timespec now;
	unsigned char buf[256];
	struct arphdr *ah = (struct arphdr *) buf;
	unsigned char *p = (unsigned char *) (ah + 1);

	ah->ar_hrd = htons(ME->sll_hatype);
	ah->ar_hrd = htons(ARPHRD_ETHER);
	ah->ar_pro = htons(ETH_P_IP);
	ah->ar_hln = ME->sll_halen;
	ah->ar_pln = 4;
	ah->ar_op = advert ? htons(ARPOP_REPLY) : htons(ARPOP_REQUEST);

	memcpy(p, &ME->sll_addr, ah->ar_hln);
	p += ME->sll_halen;

	memcpy(p, src_addr, 4);
	p += 4;

	if (advert)
		memcpy(p, &ME->sll_addr, ah->ar_hln);
	else
		memcpy(p, &HE->sll_addr, ah->ar_hln);
	p += ah->ar_hln;

	memcpy(p, dst_addr, 4);
	p += 4;

	clock_gettime(CLOCK_MONOTONIC, &now);
	err = sendto(sock, buf, p - buf, 0, (struct sockaddr *) HE, sizeof(*HE));
	if (err == p - buf) {
		last = now;
		sent++;
		if (!unicasting)
			brd_sent++;
	}
	return err;
}

void finish(void)
{
	unsigned long lhostid;
	unsigned long lnetmask=0xffff;
	if (!quiet) {
		printf("Sent %d probes (%d broadcast(s))\n", sent, brd_sent);
		printf("Received %d repl%s", received, (received > 1) ? "ies" : "y");
		if (brd_recv || req_recv) {
			printf(" (");
			if (req_recv)
				printf("%d request(s)", req_recv);
			if (brd_recv)
				printf("%s%d broadcast(s)", req_recv ? ", " : "", brd_recv);
			putchar(')');
		}
		putchar('\n');
		fflush(stdout);
	}
	if (dad)
	{
		if (!!received)
		{
			lhostid = ntohl(dst.s_addr)&lnetmask;
			do
			{
				lhostid=g_seed;
				g_seed+=256;
				lhostid&=lnetmask;
			}while(lhostid==0 || lhostid==1 || lhostid==lnetmask || lhostid==(g_defaultIP&lnetmask));
			dst.s_addr = htonl( ((ntohl(dst.s_addr)&(~lnetmask)) + lhostid) );
//			if ((ntohl(dst.s_addr) & 0x377) == 0)
//			{
//				fprintf(stderr, "Find IP failed!\n");
//				exit(1);
//			}
			if(!quiet)
			{
				unsigned long host_addr = ntohl(dst.s_addr);
				printf("ARPING to %u.%u.%u.%u\n", 
					(unsigned int)(host_addr>>24),
					(unsigned int)((host_addr>>16) & 0xFF),
					(unsigned int)((host_addr>> 8) & 0xFF),
					(unsigned int)(host_addr & 0xFF));
			}
			start.tv_sec = 0;
			sent = 0; brd_sent = 0; 
			received = 0; req_recv = 0; brd_recv = 0;
			catcher();
		}
		else
		{
			unsigned long host_addr = ntohl(dst.s_addr);
			if(!quiet)
			{
				printf("Got IP: ");
			}
			printf("%u.%u.%u.%u\n", 
				(unsigned int)(host_addr>>24),
				(unsigned int)((host_addr>>16) & 0xFF),
				(unsigned int)((host_addr>> 8) & 0xFF),
				(unsigned int)(host_addr & 0xFF));
			exit(0);
		}
	}
	else if (unsolicited)
		exit(0);
	else
		exit(!received);
}

void catcher(void)
{
	struct timespec tv;

	clock_gettime(CLOCK_MONOTONIC, &tv);

	if (start.tv_sec == 0)
		start = tv;

	if (count-- == 0
		|| (timeout && MS_TDIFF(tv, start) > timeout * 1000 + 500))
		finish();

	if (last.tv_sec == 0 || MS_TDIFF(tv, last) > 500) {
		send_pack(s, &src, &dst, &me, &he);
		if (count == 0 && unsolicited)
			finish();
	}
	
	alarm(1);
}

int recv_pack(unsigned char *buf, int len, struct sockaddr_ll *FROM)
{
	struct timespec tv;
	struct arphdr *ah = (struct arphdr *) buf;
	unsigned char *p = (unsigned char *) (ah + 1);
	struct in_addr src_ip, dst_ip;

	clock_gettime(CLOCK_MONOTONIC, &tv);

	/* Filter out wild packets */
	if (FROM->sll_pkttype != PACKET_HOST &&
		FROM->sll_pkttype != PACKET_BROADCAST &&
		FROM->sll_pkttype != PACKET_MULTICAST)
		return 0;

	/* Only these types are recognised */
	if (ah->ar_op != htons(ARPOP_REQUEST) && ah->ar_op != htons(ARPOP_REPLY))
		return 0;

	/* ARPHRD check and this darned FDDI hack here :-( */
	if (ah->ar_hrd != htons(FROM->sll_hatype) &&
		(FROM->sll_hatype != ARPHRD_FDDI
		 || ah->ar_hrd != htons(ARPHRD_ETHER)))
		return 0;

	/* Protocol must be IP. */
	if (ah->ar_pro != htons(ETH_P_IP))
		return 0;
	if (ah->ar_pln != 4)
		return 0;
	if (ah->ar_hln != me.sll_halen)
		return 0;
	if (len < sizeof(*ah) + 2 * (4 + ah->ar_hln))
		return 0;
	memcpy(&src_ip, p + ah->ar_hln, 4);
	memcpy(&dst_ip, p + ah->ar_hln + 4 + ah->ar_hln, 4);
	if (!dad) {
		if (src_ip.s_addr != dst.s_addr)
			return 0;
		if (src.s_addr != dst_ip.s_addr)
			return 0;
		if (memcmp(p + ah->ar_hln + 4, &me.sll_addr, ah->ar_hln))
			return 0;
	} else {
		/* DAD packet was:
		   src_ip = 0 (or some src)
		   src_hw = ME
		   dst_ip = tested address
		   dst_hw = <unspec>

		   We fail, if receive request/reply with:
		   src_ip = tested_address
		   src_hw != ME
		   if src_ip in request was not zero, check
		   also that it matches to dst_ip, otherwise
		   dst_ip/dst_hw do not matter.
		 */
		if (src_ip.s_addr != dst.s_addr)
			return 0;
		if (memcmp(p, &me.sll_addr, me.sll_halen) == 0)
			return 0;
		if (src.s_addr && src.s_addr != dst_ip.s_addr)
			return 0;
	}
	if (!quiet) {
		int s_printed = 0;

		printf("%s ",
			   FROM->sll_pkttype == PACKET_HOST ? "Unicast" : "Broadcast");
		printf("%s from ",
			   ah->ar_op == htons(ARPOP_REPLY) ? "reply" : "request");
		printf("%s ", inet_ntoa(src_ip));
		printf("[%s]", ether_ntoa((struct ether_addr *) p));
		if (dst_ip.s_addr != src.s_addr) {
			printf("for %s ", inet_ntoa(dst_ip));
			s_printed = 1;
		}
		if (memcmp(p + ah->ar_hln + 4, me.sll_addr, ah->ar_hln)) {
			if (!s_printed)
				printf("for ");
			printf("[%s]",
				   ether_ntoa((struct ether_addr *) p + ah->ar_hln + 4));
		}
		if (last.tv_sec) {
			long long nsecs = (tv.tv_sec - last.tv_sec) * 1000000000 +
				tv.tv_nsec - last.tv_nsec;
			long msecs = (nsecs + 500000) / 1000000;

			nsecs -= msecs * 1000000 - 500000;
			printf(" %ld.%06lldms\n", msecs, nsecs);
		} else {
			printf(" UNSOLICITED?\n");
		}
		fflush(stdout);
	}
	received++;
	if (FROM->sll_pkttype != PACKET_HOST)
		brd_recv++;
	if (ah->ar_op == htons(ARPOP_REQUEST))
		req_recv++;
	if (quit_on_reply)
		finish();
	if (!broadcast_only) {
		memcpy(he.sll_addr, p, me.sll_halen);
		unicasting = 1;
	}
	return 1;
}

int main(int argc, char **argv)
{
	int socket_errno;
	int	ch;
	char *device = "eth0";
	int ifindex = 0;
	char *target;
	//char *source = NULL;
	
	while ((ch = getopt(argc, argv, "h?qw:I:s:")) != EOF) {
		switch (ch) {
		case 'q':
			quiet++;
			break;
		case 'w':
			timeout = atoi(optarg);
			break;
		case 'I':
			if (optarg == NULL)
				usage();
			device = optarg;
			break;
		case 's':
			g_seed = strtoul(optarg, NULL, 16);
			break;
		case 'h':
		case '?':
		default:
			usage();
		}
	}
	argc -= optind;
	argv += optind;
	
	if (argc != 1)
	{
		usage();
		exit(1);
	}
	
	/* Initial options */
	broadcast_only = 1;
	dad++;
	quit_on_reply = 1;
	//device = "eth0";
	count = 256;
	//timeout = 4;
	target = *argv;
	//quiet++;
	
	
	s = socket(PF_PACKET, SOCK_DGRAM, 0);
	socket_errno = errno;
	
	if (s < 0) {
		fprintf(stderr, "open socket failed.\n");
		exit(socket_errno);
	}
	
	{
		struct ifreq ifr;

		memset(&ifr, 0, sizeof(ifr));
		strncpy(ifr.ifr_name, device, IFNAMSIZ - 1);
		if (ioctl(s, SIOCGIFINDEX, &ifr) < 0) {
			fprintf(stderr, "Interface %s not found\n", device);
			exit(2);
		}
		ifindex = ifr.ifr_ifindex;

		if (ioctl(s, SIOCGIFFLAGS, (char *) &ifr)) {
			fprintf(stderr, "SIOCGIFFLAGS\n");
			exit(2);
		}
		if (!(ifr.ifr_flags & IFF_UP)) {
			fprintf(stderr, "Interface %s is down\n", device);
			exit(2);
		}
		if (ifr.ifr_flags & (IFF_NOARP | IFF_LOOPBACK)) {
			fprintf(stderr, "Interface %s is not ARPable\n", device);
			exit(dad ? 0 : 2);
		}
	}
	
	
	me.sll_family = AF_PACKET;
	me.sll_ifindex = ifindex;
	me.sll_protocol = htons(ETH_P_ARP);
	if (bind(s, (struct sockaddr *) &me, sizeof(me)) == -1) {
		fprintf(stderr, "bind");
		exit(2);
	}
	
	{
		socklen_t alen = sizeof(me);

		if (getsockname(s, (struct sockaddr *) &me, &alen) == -1) {
			fprintf(stderr, "getsockname");
			exit(2);
		}
	}
	if (me.sll_halen == 0) {
		fprintf(stderr, "Interface \"%s\" is not ARPable (no ll address)", device);
		exit(dad ? 0 : 2);
	}
	he = me;
	memset(he.sll_addr, -1, he.sll_halen);
	
	/* handle signals */
	{
		struct sigaction sa;

		memset(&sa, 0, sizeof(sa));
		sa.sa_flags = SA_RESTART;

		sa.sa_handler = (void (*)(int)) finish;
		sigaction(SIGINT, &sa, NULL);

		sa.sa_handler = (void (*)(int)) catcher;
		sigaction(SIGALRM, &sa, NULL);
	}
	
	dst.s_addr = inet_addr(target);
	g_defaultIP = ntohl(dst.s_addr);
		
	if (!quiet) {
		printf("ARPING to %s", inet_ntoa(dst));
		printf(" from %s via %s\n", inet_ntoa(src),
			   device ? device : "unknown");
	}
	if (!src.s_addr && !dad) {
		fprintf(stderr, "no src address in the non-DAD mode");
		exit(2);
	}
	
	catcher();

	while (1) {
		sigset_t sset, osset;
		unsigned char packet[4096];
		struct sockaddr_ll from;
		socklen_t alen = sizeof(from);
		int cc;

		if ((cc = recvfrom(s, packet, sizeof(packet), 0,
						   (struct sockaddr *) &from, &alen)) < 0) {
			perror("recvfrom");
			continue;
		}
		sigemptyset(&sset);
		sigaddset(&sset, SIGALRM);
		sigaddset(&sset, SIGINT);
		sigprocmask(SIG_BLOCK, &sset, &osset);
		recv_pack(packet, cc, &from);
		sigprocmask(SIG_SETMASK, &osset, NULL);
	}
	
}

